@@ -1,15 +1,15 @@
-use crate::{
-    unicode::{BOM, WJ},
-    TextReader, TextWriter,
+//! The `TextString` and `TextStr` types.
+
+use crate::{ReadText, TextReader, TextSubstr, TextWriter};
+use basic_text_internals::{
+    is_basic_text, is_basic_text_start,
+    unicode::{BOM, CGJ, WJ},
 };
-use layered_io::{Bufferable, LayeredReader, LayeredWriter};
+use layered_io::Bufferable;
 #[cfg(try_reserve)]
 use std::collections::TryReserveError;
 #[cfg(pattern)]
-use std::str::{
-    pattern::{Pattern, ReverseSearcher},
-    MatchIndices, Matches, RMatchIndices, RMatches,
-};
+use std::str::pattern::{Pattern, ReverseSearcher};
 use std::{
     borrow::{Borrow, BorrowMut, Cow},
     cmp::Ordering,
@@ -19,35 +19,111 @@ use std::{
     hash::Hash,
     io::{self, Read, Write},
     net::{SocketAddr, ToSocketAddrs},
-    ops::{Add, AddAssign, Deref, DerefMut},
+    ops::{Add, AddAssign, Deref, DerefMut, Index, Range, RangeFrom, RangeTo},
     path::Path,
-    str::{self, Bytes, CharIndices, Chars, EncodeUtf16, FromStr, Lines, Utf8Error},
+    str::{
+        self, Bytes, CharIndices, Chars, EncodeUtf16, EscapeDebug, EscapeDefault, EscapeUnicode,
+        FromStr, Lines, MatchIndices, Matches, RMatchIndices, RMatches, Utf8Error,
+    },
     string::FromUtf8Error,
     vec,
 };
-use utf8_io::{Utf8Reader, Utf8Writer, WriteStr};
+use utf8_io::WriteStr;
 
 /// A Basic Text encoded, growable string.
+///
+/// This is an owning string similar to [`String`], but ensures the contents
+/// are Basic Text rather than just UTF-8. It's accompanied by a borrowing
+/// [`TextStr`], which plays an analogous role to [`prim@str`].
+///
+/// # Examples
+///
+/// You can create a `TextString` from a literal text string with `TextString::from`:
+///
+/// ```rust
+/// use basic_text::{text, TextString};
+///
+/// let hello = TextString::from(text!("Hello, world!"));
+/// ```
+///
+/// You can append a `&TextStr` with the `push_text` method:
+///
+/// ```rust
+/// use basic_text::{text, TextString};
+///
+/// let mut hello = TextString::from(text!("Hello, "));
+///
+/// hello.push_text(text!("world!"));
+/// ```
+///
+/// If you have a `String` containing a Basic Text string, you can create a
+/// `TextString` from it with the `from_text` method:
+///
+/// ```rust
+/// use basic_text::{text, TextString};
+///
+/// // a `String`
+/// let sparkle_heart = "💖".to_owned();
+///
+/// // We know this string is valid, so we'll use `unwrap()`.
+/// let sparkle_heart = TextString::from_text(sparkle_heart).unwrap();
+///
+/// assert_eq!(text!("💖"), &sparkle_heart);
+/// ```
+///
+/// If you have a vector of Basic Text bytes, you can create a `String` from it
+/// with the `from_text_vec` method:
+///
+/// ```rust
+/// use basic_text::{text, TextString};
+///
+/// // some bytes, in a vector
+/// let sparkle_heart = vec![240, 159, 146, 150];
+///
+/// // We know these bytes are valid, so we'll use `unwrap()`.
+/// let sparkle_heart = TextString::from_text_vec(sparkle_heart).unwrap();
+///
+/// assert_eq!(text!("💖"), &sparkle_heart);
+/// ```
 #[derive(PartialEq, Eq, Hash, Debug)]
 #[repr(transparent)]
-pub struct TextString(String);
+pub struct TextString(pub(crate) String);
 
-/// `TextStr` is to `TextString` as `str` is to `String`.
+/// Text slices.
+///
+/// `TextStr` is to `TextString` as [`prim@str`] is to `String`. It is usually
+/// used for borrowing, in the form of `&TextStr`.
+///
+/// # Examples
+///
+/// Text literals are text slices:
+///
+/// ```rust
+/// use basic_text::{text, TextStr};
+///
+/// let hello = text!("Hello, world!");
+///
+/// // with an explicit type annotation
+/// let hello: &'static TextStr = text!("Hello, world!");
+/// ```
+///
+/// They are 'static because they’re stored directly in the final binary, and so
+/// will be valid for the 'static duration.
 #[derive(PartialEq, Eq, Hash, Debug)]
 #[repr(transparent)]
-pub struct TextStr(str);
+pub struct TextStr(pub(crate) str);
 
 /// `TextError` is to `TextString` as `Utf8Error` is to `String`.
 #[derive(Copy, Clone, Eq, PartialEq, Debug)]
 pub struct TextError {
-    valid_up_to: usize,
+    pub(crate) valid_up_to: usize,
 }
 
 /// `FromTextError` is to `TextString` as `FromUtf8Error` is to `String`.
 #[derive(Clone, PartialEq, Eq, Debug)]
 pub struct FromTextError {
-    bytes: Vec<u8>,
-    error: TextError,
+    pub(crate) bytes: Vec<u8>,
+    pub(crate) error: TextError,
 }
 
 impl TextString {
@@ -75,13 +151,13 @@ impl TextString {
     #[inline]
     pub fn from_text(s: String) -> Result<Self, FromTextError> {
         let bytes: Vec<u8> = Vec::new();
-        let mut writer = TextWriter::new(Utf8Writer::new(LayeredWriter::new(bytes)));
+        let mut writer = TextWriter::new(bytes);
 
         match writer.write_str(&s).and_then(|()| writer.flush()) {
             Ok(()) => (),
             Err(_err) => {
                 writer.abandon();
-                let valid_up_to = compute_valid_string_up_to(&s);
+                let valid_up_to = compute_valid_up_to(&s);
                 return Err(FromTextError {
                     bytes: s.into_bytes(),
                     error: TextError { valid_up_to },
@@ -114,7 +190,7 @@ impl TextString {
     pub fn from_text_lossy(mut v: &str) -> Cow<TextStr> {
         // TODO: If `v` is already valid, fast-path to `Cow::Borrowed(v)`.
         // TODO: Also, this currently redoes UTF-8 validation for `v`.
-        let mut reader = TextReader::new(Utf8Reader::new(LayeredReader::new(v.as_bytes())));
+        let mut reader = TextReader::new(v.as_bytes());
         let mut text = String::new();
         reader.read_to_string(&mut text).unwrap();
 
@@ -170,7 +246,7 @@ impl TextString {
     /// Converts a `TextString` into a `String`.
     #[inline]
     #[must_use]
-    pub fn into_utf8(self) -> String {
+    pub fn into_string(self) -> String {
         self.0
     }
 
@@ -184,38 +260,37 @@ impl TextString {
     /// Extracts a UTF-8 string slice containing the entire `TextString`.
     #[inline]
     #[must_use]
-    pub fn as_utf8(&self) -> &str {
+    pub fn as_str(&self) -> &str {
         &self.0
     }
 
     /// Extracts a Basic Text string slice containing the entire `TextString`.
     #[inline]
     #[must_use]
-    pub fn as_str(&self) -> &TextStr {
+    pub fn as_text(&self) -> &TextStr {
         self
     }
 
-    /// Converts a `TextString` into a mutable UTF-8 string slice.
-    #[inline]
-    #[must_use]
-    pub fn as_mut_utf8(&mut self) -> &mut str {
-        &mut self.0
-    }
-
     /// Converts a `TextString` into a mutable Basic Text string slice.
     #[inline]
     #[must_use]
-    pub fn as_mut_str(&mut self) -> &mut TextStr {
+    pub fn as_mut_text(&mut self) -> &mut TextStr {
         self
     }
 
     /// Appends a given string slice onto the end of this `TextString`.
     ///
-    /// But wait, NFC isn't closed under concatenation! This is true, but
+    /// But wait, isn't NFC closed under concatenation? This is true, but
     /// Basic Text has additional restrictions, including that strings start
-    /// with non-combining codepoints, so it is closed under concatenation.
+    /// with non-combining codepoints, so it *is* closed under concatenation.
     #[inline]
-    pub fn push_str(&mut self, s: &TextStr) {
+    pub fn push_text(&mut self, s: &TextStr) {
+        if let Some(rest) = s.0.strip_prefix(CGJ) {
+            if self.0.ends_with(CGJ) {
+                self.0.push_str(rest);
+                return;
+            }
+        }
         self.0.push_str(&s.0);
     }
 
@@ -293,6 +368,19 @@ impl TextString {
         self.0.as_mut_vec()
     }
 
+    /// Returns a mutable reference to the contents of this `TextString`.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe because it does not check that the bytes passed
+    /// to it are valid Basic Text. If this constraint is violated, it may
+    /// cause memory unsafety issues with future users of the String, as the
+    /// rest of this crate assumes that `TextString`s are valid Basic Text.
+    #[inline]
+    pub unsafe fn as_mut_string(&mut self) -> &mut String {
+        &mut self.0
+    }
+
     /// Returns the length of this `TextString`, in bytes, not `char`s or
     /// graphemes.
     #[inline]
@@ -319,20 +407,20 @@ impl TextString {
 
     /// Converts this `TextString` into a `Box<str>`.
     #[inline]
-    pub fn into_boxed_utf8(self) -> Box<str> {
+    pub fn into_boxed_str(self) -> Box<str> {
         self.0.into_boxed_str()
     }
 
     /// Converts this `TextString` into a `Box<TextStr>`.
     #[inline]
-    pub fn into_boxed_str(self) -> Box<TextStr> {
-        let slice = self.into_boxed_utf8();
+    pub fn into_boxed_text(self) -> Box<TextStr> {
+        let slice = self.into_boxed_str();
         unsafe { TextStr::from_boxed_text_unchecked(slice) }
     }
 }
 
 #[cold]
-fn compute_valid_str_up_to(s: &str) -> usize {
+fn compute_valid_up_to(s: &str) -> usize {
     // Binary search in `s` for the place where the error starts. We do
     // this after the fact rather than tracking the positions of everything
     // as we go, because tracking the positions through multiple iterators
@@ -354,47 +442,15 @@ fn compute_valid_str_up_to(s: &str) -> usize {
             }
         }
         let substr = &s[..mid];
-        match TextString::from_text(substr.to_owned()) {
-            Ok(text_string) if text_string.as_utf8() == substr => begin = mid,
-            _ => end = mid,
+        if is_basic_text(substr) {
+            begin = mid;
+        } else {
+            end = mid;
         }
     }
     begin
 }
 
-#[cold]
-fn compute_valid_string_up_to(s: &String) -> usize {
-    // Binary search in `s` for the place where the error starts. We do
-    // this after the fact rather than tracking the positions of everything
-    // as we go, because tracking the positions through multiple iterators
-    // is complex.
-    let mut begin = 0;
-    let mut end = s.len();
-    while begin != end {
-        let bytes = Vec::new();
-        let mut writer = TextWriter::new(Utf8Writer::new(LayeredWriter::new(bytes)));
-        let mut mid = begin + (end - begin) / 2;
-        while !s.is_char_boundary(mid) {
-            mid -= 1;
-        }
-        if mid == begin {
-            mid = begin + (end - begin) / 2 + 1;
-            while !s.is_char_boundary(mid) {
-                mid += 1;
-            }
-            if mid == end {
-                break;
-            }
-        }
-        match writer.write_str(&s[..mid]).and_then(|()| writer.flush()) {
-            Ok(()) => begin = mid,
-            Err(_err) => end = mid,
-        }
-        writer.abandon();
-    }
-    begin
-}
-
 impl AsRef<[u8]> for TextString {
     #[inline]
     fn as_ref(&self) -> &[u8] {
@@ -421,7 +477,7 @@ impl AsRef<Path> for TextString {
 impl AsRef<str> for TextString {
     #[inline]
     fn as_ref(&self) -> &str {
-        self.as_utf8()
+        self.as_str()
     }
 }
 
@@ -432,6 +488,13 @@ impl AsRef<TextStr> for TextString {
     }
 }
 
+impl AsRef<TextSubstr> for TextString {
+    #[inline]
+    fn as_ref(&self) -> &TextSubstr {
+        unsafe { TextSubstr::from_text_unchecked(&self.0) }
+    }
+}
+
 impl Clone for TextString {
     #[inline]
     fn clone(&self) -> Self {
@@ -488,7 +551,7 @@ impl Add<&TextStr> for TextString {
 
     #[inline]
     fn add(mut self, other: &TextStr) -> Self::Output {
-        self.push_str(other);
+        self.push_text(other);
         self
     }
 }
@@ -496,61 +559,46 @@ impl Add<&TextStr> for TextString {
 impl AddAssign<&TextStr> for TextString {
     #[inline]
     fn add_assign(&mut self, other: &TextStr) {
-        self.push_str(other);
+        self.push_text(other);
     }
 }
 
-impl<'a> PartialEq<&'a TextStr> for TextString {
+impl Extend<TextString> for TextString {
+    fn extend<I: IntoIterator<Item = TextString>>(&mut self, iter: I) {
+        iter.into_iter().for_each(move |s| self.push_text(&s));
+    }
+
+    #[cfg(extend_one)]
     #[inline]
-    fn eq(&self, other: &&'a TextStr) -> bool {
-        self.0.eq(&other.0)
+    fn extend_one(&mut self, s: TextString) {
+        self.push_text(&s);
     }
 }
 
-impl<'a> PartialEq<Cow<'a, TextStr>> for TextString {
+impl<'a> Extend<&'a TextStr> for TextString {
+    fn extend<I: IntoIterator<Item = &'a TextStr>>(&mut self, iter: I) {
+        iter.into_iter().for_each(move |s| self.push_text(s));
+    }
+
+    #[cfg(extend_one)]
     #[inline]
-    fn eq(&self, other: &Cow<'a, TextStr>) -> bool {
-        self.0.eq(&other.0)
+    fn extend_one(&mut self, s: &'a TextStr) {
+        self.push_text(s);
     }
 }
 
-impl<'a> PartialEq<TextString> for Cow<'a, TextStr> {
+impl<'a> Extend<Cow<'a, TextStr>> for TextString {
+    fn extend<I: IntoIterator<Item = Cow<'a, TextStr>>>(&mut self, iter: I) {
+        iter.into_iter().for_each(move |s| self.push_text(&s));
+    }
+
+    #[cfg(extend_one)]
     #[inline]
-    fn eq(&self, other: &TextString) -> bool {
-        self.0.eq(&other.0)
+    fn extend_one(&mut self, s: Cow<'a, TextStr>) {
+        self.push_text(&s);
     }
 }
 
-impl<'a> PartialEq<str> for TextString {
-    #[inline]
-    fn eq(&self, other: &str) -> bool {
-        self.0.eq(other)
-    }
-}
-
-impl<'a> PartialEq<String> for TextString {
-    #[inline]
-    fn eq(&self, other: &String) -> bool {
-        self.0.eq(other)
-    }
-}
-
-impl<'a> PartialEq<TextString> for String {
-    #[inline]
-    fn eq(&self, other: &TextString) -> bool {
-        self.eq(&other.0)
-    }
-}
-
-impl<'a> PartialEq<TextString> for &'a str {
-    #[inline]
-    fn eq(&self, other: &TextString) -> bool {
-        self.eq(&other.0)
-    }
-}
-
-// TODO: impl Extend for TextString?
-
 impl TextStr {
     /// Converts a slice of bytes to a text string slice.
     #[inline]
@@ -561,12 +609,8 @@ impl TextStr {
     /// Converts a string slice to a text string slice.
     #[inline]
     pub fn from_text(s: &str) -> Result<&Self, TextError> {
-        // TODO: Do this without constructing temporaries.
-        let text_string = TextString::from_text(s.to_owned()).map_err(|e| e.text_error())?;
-
-        // If any bytes got rewritten (eg. by normalization), fail.
-        if text_string.as_utf8() != s {
-            let valid_up_to = compute_valid_str_up_to(s);
+        if !is_basic_text(s) {
+            let valid_up_to = compute_valid_up_to(s);
             return Err(TextError { valid_up_to });
         }
 
@@ -582,12 +626,8 @@ impl TextStr {
     /// Converts a mutable string slice to a mutable text string slice.
     #[inline]
     pub fn from_text_mut(s: &mut str) -> Result<&mut Self, TextError> {
-        // TODO: Do this without constructing temporaries.
-        let text_string = TextString::from_text(s.to_owned()).map_err(|e| e.text_error())?;
-
-        // If any bytes got rewritten (eg. by normalization), fail.
-        if text_string.as_utf8() != s {
-            let valid_up_to = compute_valid_str_up_to(s);
+        if !is_basic_text(s) {
+            let valid_up_to = compute_valid_up_to(s);
             return Err(TextError { valid_up_to });
         }
 
@@ -730,7 +770,7 @@ impl TextStr {
 
     /// Extracts a UTF-8 string slice containing the entire `TextStr`.
     #[inline]
-    pub fn as_utf8(&self) -> &str {
+    pub fn as_str(&self) -> &str {
         &self.0
     }
 
@@ -756,7 +796,7 @@ impl TextStr {
         )
     }
 
-    // TODO: get*, slice*
+    // TODO: get*
 
     /// Returns an iterator over the `char`s of a text string slice.
     #[inline]
@@ -780,6 +820,8 @@ impl TextStr {
     // TODO: split*?
 
     /// An iterator over the lines of a text string, as text string slices.
+    ///
+    /// TODO: There should be a `TextLines` which yields `&TextStr`s.
     #[inline]
     pub fn lines(&self) -> Lines {
         self.0.lines()
@@ -804,6 +846,16 @@ impl TextStr {
         self.0.contains(pat)
     }
 
+    /// Returns `true` if the given pattern matches a sub-slice of this
+    /// text string slice.
+    ///
+    /// Returns `false` if it does not.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn contains<'a>(&'a self, pat: &str) -> bool {
+        self.0.contains(pat)
+    }
+
     /// Returns `true` if the given pattern matches a prefix of this
     /// text string slice.
     ///
@@ -817,6 +869,16 @@ impl TextStr {
         self.0.starts_with(pat)
     }
 
+    /// Returns `true` if the given pattern matches a prefix of this
+    /// text string slice.
+    ///
+    /// Returns `false` if it does not.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn starts_with<'a>(&'a self, pat: &str) -> bool {
+        self.0.starts_with(pat)
+    }
+
     /// Returns `true` if the given pattern matches a suffix of this
     /// text string slice.
     ///
@@ -831,6 +893,16 @@ impl TextStr {
         self.0.ends_with(pat)
     }
 
+    /// Returns `true` if the given pattern matches a suffix of this
+    /// text string slice.
+    ///
+    /// Returns `false` if it does not.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn ends_with<'a>(&'a self, pat: &str) -> bool {
+        self.0.ends_with(pat)
+    }
+
     /// Returns the byte index of the first character of this text string slice
     /// that matches the pattern.
     ///
@@ -844,6 +916,16 @@ impl TextStr {
         self.0.find(pat)
     }
 
+    /// Returns the byte index of the first character of this text string slice
+    /// that matches the pattern.
+    ///
+    /// Returns `None` if the pattern doesn't match.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn find<'a>(&'a self, pat: &str) -> Option<usize> {
+        self.0.find(pat)
+    }
+
     /// Returns the byte index for the first character of the rightmost match of
     /// the pattern in this text string slice.
     ///
@@ -858,6 +940,16 @@ impl TextStr {
         self.0.rfind(pat)
     }
 
+    /// Returns the byte index for the first character of the rightmost match of
+    /// the pattern in this text string slice.
+    ///
+    /// Returns `None` if the pattern doesn't match.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn rfind<'a>(&'a self, pat: &str) -> Option<usize> {
+        self.0.rfind(pat)
+    }
+
     // TODO: *split*?
 
     /// An iterator over the disjoint matches of a pattern within the given
@@ -871,8 +963,20 @@ impl TextStr {
         self.0.matches(pat)
     }
 
+    /// An iterator over the disjoint matches of a pattern within the given
+    /// text string slice.
+    ///
+    /// TODO: There should be a `TextMatches` which yields `&TextStr`s.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn matches<'a>(&'a self, pat: &'a str) -> Matches<'a, &str> {
+        self.0.matches(pat)
+    }
+
     /// An iterator over the disjoint matches of a pattern within this
     /// text string slice, yielded in reverse order.
+    ///
+    /// TODO: There should be a `TextRMatches` which yields `&TextStr`s.
     #[cfg(pattern)]
     #[inline]
     pub fn rmatches<'a, P>(&'a self, pat: P) -> RMatches<'a, P>
@@ -883,8 +987,18 @@ impl TextStr {
         self.0.rmatches(pat)
     }
 
+    /// An iterator over the disjoint matches of a pattern within this
+    /// text string slice, yielded in reverse order.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn rmatches<'a>(&'a self, pat: &'a str) -> RMatches<'a, &'a str> {
+        self.0.rmatches(pat)
+    }
+
     /// An iterator over the disjoint matches of a pattern within this
     /// text string slice as well as the index that the match starts at.
+    ///
+    /// TODO: There should be a `TextMatchIndices` which yields `&TextStr`s.
     #[cfg(pattern)]
     #[inline]
     pub fn match_indices<'a, P>(&'a self, pat: P) -> MatchIndices<'a, P>
@@ -894,8 +1008,18 @@ impl TextStr {
         self.0.match_indices(pat)
     }
 
+    /// An iterator over the disjoint matches of a pattern within this
+    /// text string slice as well as the index that the match starts at.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn match_indices<'a>(&'a self, pat: &'a str) -> MatchIndices<'a, &'a str> {
+        self.0.match_indices(pat)
+    }
+
     /// An iterator over the disjoint matches of a pattern within `self`,
     /// yielded in reverse order along with the index of the match.
+    ///
+    /// TODO: There should be a `TextRMatchIndices` which yields `&TextStr`s.
     #[cfg(pattern)]
     #[inline]
     pub fn rmatch_indices<'a, P>(&'a self, pat: P) -> RMatchIndices<'a, P>
@@ -906,6 +1030,14 @@ impl TextStr {
         self.0.rmatch_indices(pat)
     }
 
+    /// An iterator over the disjoint matches of a pattern within `self`,
+    /// yielded in reverse order along with the index of the match.
+    #[cfg(not(pattern))]
+    #[inline]
+    pub fn rmatch_indices<'a>(&'a self, pat: &'a str) -> RMatchIndices<'a, &'a str> {
+        self.0.rmatch_indices(pat)
+    }
+
     /// Returns a text string slice with leading and trailing whitespace removed.
     #[inline]
     pub fn trim(&self) -> &Self {
@@ -954,19 +1086,19 @@ impl TextStr {
     /// allocating.
     #[inline]
     pub fn into_boxed_bytes(self: Box<Self>) -> Box<[u8]> {
-        self.into_boxed_str().into_boxed_bytes()
+        self.into_boxed_text().into_boxed_bytes()
     }
 
     /// Converts a `Box<TextStr>` into a `Box<str>` without copying or
     /// allocating.
     #[inline]
-    pub fn into_boxed_str(self: Box<Self>) -> Box<str> {
+    pub fn into_boxed_text(self: Box<Self>) -> Box<str> {
         self.into()
     }
 
     /// Converts a `Box<TextStr>` into a `String` without copying or allocating.
     #[inline]
-    pub fn into_utf8(self: Box<Self>) -> String {
+    pub fn into_string(self: Box<Self>) -> String {
         let slice = Box::<[u8]>::from(self);
         unsafe { String::from_utf8_unchecked(slice.into_vec()) }
     }
@@ -974,14 +1106,39 @@ impl TextStr {
     /// Converts a `Box<TextStr>` into a `TextString` without copying or
     /// allocating.
     #[inline]
-    pub fn into_string(self: Box<Self>) -> TextString {
-        unsafe { TextString::from_text_unchecked(Self::into_utf8(self)) }
+    pub fn into_text_string(self: Box<Self>) -> TextString {
+        unsafe { TextString::from_text_unchecked(Self::into_string(self)) }
     }
 
-    // TODO: make_ascii_uppercase, make_ascii_lowercase, escape_debug,
-    // escape_default, escape_unicode, replace*, to_lowercase, to_uppercase,
-    // into_string, repeat, to_ascii_uppercase, to_ascii_lowercase; determine
-    // whether these can be done without breaking NFC.
+    /// Creates a new [`TextString`] by repeating a string `n` times.
+    pub fn repeat(&self, n: usize) -> TextString {
+        unsafe { TextString::from_text_vec_unchecked(self.as_bytes().repeat(n)) }
+    }
+
+    /// Return an iterator that escapes each `char` in `self` with
+    /// [`char::escape_debug`].
+    #[inline]
+    pub fn escape_debug(&self) -> EscapeDebug<'_> {
+        self.0.escape_debug()
+    }
+
+    /// Return an iterator that escapes each `char` in `self` with
+    /// [`char::escape_default`].
+    #[inline]
+    pub fn escape_default(&self) -> EscapeDefault<'_> {
+        self.0.escape_default()
+    }
+
+    /// Return an iterator that escapes each `char` in `self` with
+    /// [`char::escape_unicode`].
+    #[inline]
+    pub fn escape_unicode(&self) -> EscapeUnicode {
+        self.0.escape_unicode()
+    }
+
+    // TODO: make_ascii_uppercase, make_ascii_lowercase, replace*,
+    // to_lowercase, to_uppercase, to_ascii_uppercase, to_ascii_lowercase;
+    // determine whether these can be done without breaking NFC.
 }
 
 impl AsRef<[u8]> for TextStr {
@@ -1019,6 +1176,13 @@ impl AsRef<Self> for TextStr {
     }
 }
 
+impl AsRef<TextSubstr> for TextStr {
+    #[inline]
+    fn as_ref(&self) -> &TextSubstr {
+        unsafe { TextSubstr::from_text_unchecked(&self.0) }
+    }
+}
+
 impl Default for &TextStr {
     #[inline]
     fn default() -> Self {
@@ -1034,13 +1198,11 @@ impl Default for &mut TextStr {
 }
 
 impl Display for TextStr {
-    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         Display::fmt(&self.0, f)
     }
 }
 
-// TODO: impl Index* for TextStr?
-
 impl Ord for TextStr {
     #[inline]
     fn cmp(&self, other: &Self) -> Ordering {
@@ -1048,43 +1210,6 @@ impl Ord for TextStr {
     }
 }
 
-impl<'a> PartialEq<Cow<'a, Self>> for TextStr {
-    #[inline]
-    fn eq(&self, other: &Cow<'a, Self>) -> bool {
-        self.0.eq(&other.0)
-    }
-}
-
-impl<'a, 'b> PartialEq<Cow<'a, TextStr>> for &'b TextStr {
-    #[inline]
-    fn eq(&self, other: &Cow<'a, TextStr>) -> bool {
-        self.0.eq(&other.0)
-    }
-}
-
-impl<'a> PartialEq<str> for TextStr {
-    #[inline]
-    fn eq(&self, other: &str) -> bool {
-        self.0.eq(other)
-    }
-}
-
-impl<'a> PartialEq<String> for TextStr {
-    #[inline]
-    fn eq(&self, other: &String) -> bool {
-        self.0.eq(other)
-    }
-}
-
-impl<'a> PartialEq<TextStr> for String {
-    #[inline]
-    fn eq(&self, other: &TextStr) -> bool {
-        self.eq(&other.0)
-    }
-}
-
-// TODO: all the PartialEq impls for TextStr
-
 impl PartialOrd<Self> for TextStr {
     #[inline]
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
@@ -1092,6 +1217,13 @@ impl PartialOrd<Self> for TextStr {
     }
 }
 
+impl PartialOrd<Self> for TextString {
+    #[inline]
+    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
+        self.0.partial_cmp(&other.0)
+    }
+}
+
 // TODO: Pattern for TextStr
 
 impl ToOwned for TextStr {
@@ -1113,8 +1245,8 @@ impl ToSocketAddrs for TextStr {
 }
 
 impl TextError {
-    // Returns the index in the given string up to which valid Basic Text was
-    // verified.
+    /// Returns the index in the given string up to which valid Basic Text was
+    /// verified.
     pub fn valid_up_to(&self) -> usize {
         self.valid_up_to
     }
@@ -1129,8 +1261,12 @@ impl From<Utf8Error> for TextError {
 }
 
 impl Display for TextError {
-    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
-        write!(f, "TODO: Display for TextError: {:?}", self)
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "invalid Basic Text byte sequence from index {}",
+            self.valid_up_to
+        )
     }
 }
 
@@ -1167,8 +1303,8 @@ impl From<FromUtf8Error> for FromTextError {
 }
 
 impl Display for FromTextError {
-    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
-        write!(f, "TODO: Display for FromTextError: {:?}", self)
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        fmt::Display::fmt(&self.error, f)
     }
 }
 
@@ -1193,7 +1329,7 @@ impl From<Box<TextStr>> for Box<str> {
 impl From<Box<TextStr>> for TextString {
     #[inline]
     fn from(s: Box<TextStr>) -> Self {
-        s.into_string()
+        s.into_text_string()
     }
 }
 
@@ -1231,7 +1367,7 @@ impl From<Cow<'_, TextStr>> for Box<TextStr> {
 impl From<TextString> for Box<TextStr> {
     #[inline]
     fn from(s: TextString) -> Self {
-        s.into_boxed_str()
+        s.into_boxed_text()
     }
 }
 
@@ -1307,9 +1443,82 @@ impl<'a> arbitrary::Arbitrary<'a> for TextString {
     }
 }
 
+impl Index<Range<usize>> for TextString {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: Range<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+impl Index<Range<usize>> for TextStr {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: Range<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+impl Index<RangeTo<usize>> for TextString {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: RangeTo<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+impl Index<RangeTo<usize>> for TextStr {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: RangeTo<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+impl Index<RangeFrom<usize>> for TextString {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: RangeFrom<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+impl Index<RangeFrom<usize>> for TextStr {
+    type Output = TextSubstr;
+
+    #[inline]
+    fn index(&self, index: RangeFrom<usize>) -> &Self::Output {
+        unsafe { TextSubstr::from_text_unchecked(self.0.index(index)) }
+    }
+}
+
+/// Default implementation of [`ReadText::read_to_text_string`].
+pub fn default_read_to_text_string<Inner: ReadText + ?Sized>(
+    inner: &mut Inner,
+    buf: &mut TextString,
+) -> io::Result<usize> {
+    // Read directly into the inner `String`.
+    let start = buf.0.len();
+    let n = inner.read_to_string(&mut buf.0)?;
+    if let Some(c) = buf.0[start..].chars().next() {
+        if !is_basic_text_start(c) {
+            return Err(io::Error::new(
+                io::ErrorKind::InvalidData,
+                "read_to_text_string requires a starter",
+            ));
+        }
+    }
+    Ok(n)
+}
+
 #[test]
 fn normalize_string() {
-    let ring = "\u{030a}";
+    let ring = "\u{30a}";
     let unnormal = "A\u{30a}";
     let unnormal_nl = "A\u{30a}\n";
     let composed = TextStr::from_text("\u{c5}").unwrap();
